#include <memory>

#include "logcategories.h"
#include "subdivision.h"

Subdivision::Subdivision(QOpenGLFunctions_4_3_Core *f)
{
    this->f =f;
}

Subdivision::~Subdivision()
{
    delete edgeShader;
}

QString Subdivision::formatTimeMeasurement(int time){
   return QString("%1m %2s %3ms").arg( time / 60000         , 2, 10, QChar('0'))
                                  .arg((time % 60000) / 1000, 2, 10, QChar('0'))
                                   .arg((time % 1000)       , 3, 10, QChar('0'));
}

void Subdivision::init() {
    QString source = QLatin1String(":/subdivision-edge.compute");
    edgeShader = initComputeShaderProgram(source);

    source = QLatin1String(":/subdivision-vertex.compute");
    vertexShader = initComputeShaderProgram(source);
}

QOpenGLShaderProgram *Subdivision::initComputeShaderProgram(QString &source){
    qCDebug(log_subdiv) << "Compiling compute shader ...";
    QOpenGLShader *computeShader = new QOpenGLShader(QOpenGLShader::Compute);
    if(!computeShader->compileSourceFile(source)){
        qCCritical(log_subdiv) << "Compute Shader" << source << "failed" << computeShader->log();
        exit(5);
    }

    qCDebug(log_subdiv) << "Adding compute shader ...";
    QOpenGLShaderProgram *shader = new QOpenGLShaderProgram();
    shader->addShader(computeShader);
    qCDebug(log_subdiv) << "Linking compute shader ...";
    if(!shader->link()){
        qCCritical(log_subdiv) << "Linking compute shader failed:" << shader->log();
        exit(5);
    }
    qCDebug(log_subdiv) << "Linking compute shader done";

    return shader;
}

void Subdivision::subdivide(Mesh *mesh, int level) {
    // For now we only look at the first mesh entry
    Mesh::Node root = mesh->getRootNode();

    int first_mesh_index = -1;
    if (!root.getFirstMeshIndex(first_mesh_index)) {
        qCCritical(log_subdiv) << "No mesh found, aborting subdivision";
        return;
    }


    Mesh::MeshEntry *current_mesh = mesh->getMeshEntry(first_mesh_index);
    Input input;

    int currentMax = current_mesh->buffers.size()-1;
    //qCDebug(log_subdiv) << "Subdiv Level" << level << "Requested. Current Max:" << currentMax;
    if(level <= currentMax)
        return;

    std::shared_ptr<Mesh::SubdivEntry> entry = current_mesh->buffers[currentMax];

    input.vb_handle = entry->VB_handle;
    input.vertex_buffer = entry->vertices;
    input.index_irregular_buffer = entry->indices_irregular;
    input.index_regular_buffer = patchIBToTriangleIB(entry->indices_regular);

    if (input.index_irregular_buffer.isEmpty()) {
        //  current_mesh->update(input.vb_handle, input.vertex_buffer, input.index_buffer, entry->indices_regular);
    } else {
        Tables tables = precomputeTables(input);
        Result result = runShader(input, tables);

        QVector<unsigned int> regular;
        QVector<unsigned int> irregular;
        findRegular(tables.index_buffer, result.vertex_buffer, regular, irregular);

        qCDebug(log_subdiv) << "Indices" << tables.index_buffer.length();
        qCDebug(log_subdiv) << "subdivide: regular" << regular.length();
        qCDebug(log_subdiv) << "subdivide: irregular" << irregular.length();

        QVector<unsigned int> patches = getPatchIndexBuffer(regular, irregular,result.vertex_buffer);//todo is this the correct vb?
        qCDebug(log_subdiv) << "patches" << patches.length();

        current_mesh->update(result.vb_handle, result.vertex_buffer, irregular, patches);
    }
}

void insert_edge(QMap<Triangle, QVector<Triangle::Edge>> &opposite_edges, const Triangle &triangle, const Triangle::Edge &edge_a, const Triangle::Edge &edge_b) {
    QVector<Triangle::Edge> edges = opposite_edges.value(triangle, QVector<Triangle::Edge>());
    edges.resize(3);
    switch (edge_a.name) {
    case Triangle::Edge::Name::uv:
        edges[0] = edge_b;
        break;
    case Triangle::Edge::Name::vw:
        edges[1] = edge_b;
        break;
    case Triangle::Edge::Name::wu:
        edges[2] = edge_b;
        break;
    default:
        qCWarning(log_subdiv) << "got" << edge_a.name << "as edge!";
        break;
    }
    opposite_edges.insert(triangle, edges);
}

struct Edge {
    unsigned int a;
    unsigned int b;

    friend bool operator<(const Edge &l, const Edge &r) {
        if (l.a < r.a) {
            return true;
        } else if (l.a > r.a) {
            return false;
        } else {
            return l.b < r.b;
        }
    }
};

/**
 * @brief Subdivision::precomputeTables
 * @param mesh
 * @param edgeIndices_base Wird gefüllt mit 4er-Arrays, die die indices in den alten Vertexbuffer enthalten, von denen der jeweilige edge-point abhängt
 * @param vertexIndices Wird gefüllt mit Folgen von ints, die die indices in den alten Vertexbuffer enthalten, von denen die neuen vertex-points abhängen
 * @param vertexIndicesOffsets enthält offsets in vertexIndices
 * @param newIndexBuffer Wird gefüllt mit neuem Indexbuffer. Der Indexbuffer geht davon aus, dass im neuen Vertex-Buffer
 * erst die Vertex-points (in der selben Reihenfolge wie im ursprünglichen Vertex-Buffer) und dann die Edgepoints kommen.
 *
 * Einträge in neuem Vertex-buffer: Verschobene Knoten aus altem Vertex-Buffer :: edge-vertices erstellt aus edgeIndices_base.
 */
Subdivision::Tables Subdivision::precomputeTables(Input input) {
    Tables tables;

    QVector<unsigned int> ib = input.index_irregular_buffer;
    qCDebug(log_subdiv_trace) << "Index Buffer Irregular: " << ib;

    QVector<unsigned int> ib_regular = input.index_regular_buffer;
    qCDebug(log_subdiv_trace) << "Index Buffer Regular: " << ib_regular;

    QVector<Vertex> vb = input.vertex_buffer;
    qCDebug(log_subdiv_trace) << "Vertex Buffer: " << vb;

    QTime trianglesTimer;
    trianglesTimer.start();
    QVector<Triangle> triangles;
    for (int i = 0; i < ib.length(); i+=3) {
        triangles.push_back(Triangle(&vb, ib[i], ib[i+1], ib[i+2]));
    }
    QVector<Triangle> triangles_regular;
    for (int i = 0; i < ib_regular.length(); i+=3) {
        triangles.push_back(Triangle(&vb, ib_regular[i], ib_regular[i+1], ib_regular[i+2]));
    }
    qCDebug(log_timing) << "building Triangles:" << formatTimeMeasurement(trianglesTimer.elapsed());

    QTime subTimer;

    subTimer.start();
    precomputeEdgeTable(tables, triangles, triangles_regular, (unsigned int) vb.length());
    qCDebug(log_timing) << "precomputeEdgeTable:" << formatTimeMeasurement(subTimer.elapsed());

    subTimer.restart();
    precomputeVertexTable(tables, input);
    qCDebug(log_timing) << "precomputeVertexTable:" << formatTimeMeasurement(subTimer.elapsed());

    //tables.patch_index_regular = getPatchIndexBuffer(ib, QVector<unsigned int>());//TODO pass regular and irregular index buffer?
    qCDebug(log_subdiv) << "Precompute Tables Done";
    return tables;
}

void Subdivision::precomputeEdgeTable(Subdivision::Tables &tables, QVector<Triangle> &triangles, QVector<Triangle> &triangles_regular, unsigned int offset) {
    //compute edge table
    //Format: first two entries: edge vertices. last two entries: distant vertices.

    // for each triangle the edges of the neighbor triangles are stored.
    // First the edge shared with the triangles uv edge, then vw and wu.
    QMap<Triangle, QVector<Triangle::Edge>> opposite_edges;

    // Maps edges to the index of the new vertex added on that edge. The keys are the indices of the two vertices defining the edge and must be in the order uv, vw or uw.
    QMap<Edge, unsigned int> edge_indices;

    for (int i = 0; i < triangles.length(); i++){
        //schaue alle dreiecke an
        Triangle triangle = triangles[i];

        QVector<unsigned int> edge_indices_buffer;//push all edge indices into this, then check if enough edge indices were found. if yes, copy to Tables and add to new index buffer.

        /*
         * Find edge xy in all other triangles, add to edge_indices
         * Find edge yz
         * Find edge zx
         * - We assume that we have a closed surface, i.e. each triangle has exactly one adjacent triangle at each edge
         * - Each edge point will be computed twice. This allows us to implicitly know the adjacent points.
         */

        Triangle::Edge edge_a, edge_b;
        for (int j = i + 1; j < triangles.length(); j++){
            if (triangle.get_shared_edge(triangles[j], edge_a, edge_b)) {
                insert_edge(opposite_edges, triangle, edge_a, edge_b);
                insert_edge(opposite_edges, triangles[j], edge_b, edge_a);
            }
        }
        for (int j = 0; j < triangles_regular.length(); j++){
            if (triangle.get_shared_edge(triangles_regular[j], edge_a, edge_b)) {
                insert_edge(opposite_edges, triangle, edge_a, edge_b);
                // no need to insert the opposite edge of a regular triangle - it won't be processed by the outer loop
            }
        }

        QVector<Triangle::Edge> opposite = opposite_edges.value(triangle);

        // indices of the three vertices added to the edges of this triangle
        unsigned int uv, vw, wu;

        Edge edge;

        edge = { triangle.u_idx(), triangle.v_idx() };
        if (edge_indices.contains(edge)) {
            uv = edge_indices.value(edge);
        } else {
            uv = offset++;
            edge_indices.insert(edge, uv);

            edge = { opposite[0].a, opposite[0].b };
            edge_indices.insert(edge, uv);

            tables.edge_indices.push_back(triangle.u_idx());
            tables.edge_indices.push_back(triangle.v_idx());
            tables.edge_indices.push_back(triangle.w_idx());
            tables.edge_indices.push_back(opposite[0].c);
        }

        edge = { triangle.v_idx(), triangle.w_idx() };
        if (edge_indices.contains(edge)) {
            vw = edge_indices.value(edge);
        } else {
            vw = offset++;
            edge_indices.insert(edge, vw);

            edge = { opposite[1].a, opposite[1].b };
            edge_indices.insert(edge, vw);

            tables.edge_indices.push_back(triangle.v_idx());
            tables.edge_indices.push_back(triangle.w_idx());
            tables.edge_indices.push_back(triangle.u_idx());
            tables.edge_indices.push_back(opposite[1].c);
        }

        edge = { triangle.w_idx(), triangle.u_idx() };
        if (edge_indices.contains(edge)) {
            wu = edge_indices.value(edge);
        } else {
            wu = offset++;
            edge_indices.insert(edge, wu);

            edge = { opposite[2].a, opposite[2].b };
            edge_indices.insert(edge, wu);

            tables.edge_indices.push_back(triangle.w_idx());
            tables.edge_indices.push_back(triangle.u_idx());
            tables.edge_indices.push_back(triangle.v_idx());
            tables.edge_indices.push_back(opposite[2].c);
        }

        //add indices to new index buffer
        tables.index_buffer.push_back(triangle.u_idx());
        tables.index_buffer.push_back(uv);
        tables.index_buffer.push_back(wu);

        tables.index_buffer.push_back(triangle.v_idx());
        tables.index_buffer.push_back(vw);
        tables.index_buffer.push_back(uv);

        tables.index_buffer.push_back(triangle.w_idx());
        tables.index_buffer.push_back(wu);
        tables.index_buffer.push_back(vw);

        tables.index_buffer.push_back(uv);
        tables.index_buffer.push_back(vw);
        tables.index_buffer.push_back(wu);
    }

    qCDebug(log_subdiv) << "Done with edge table. " << tables.edge_indices.length();
    if (log_subdiv_trace().isDebugEnabled()) {
        qCDebug(log_subdiv) << "Eedges found. Table: " << tables.edge_indices;
        qCDebug(log_subdiv) << "Table (long version):";
        for (int i = 0; i < tables.edge_indices.length(); i++){
            qCDebug(log_subdiv) << "blub:" << tables.edge_indices[i];
        }
    }
}

void Subdivision::precomputeVertexTable(Subdivision::Tables &tables, Input &input) {
    //compute vertex table
    //Format: First entry: original vertex. Other entries: adjacent vertices.
    QVector<QVector<unsigned int> > duplicates;//for debugging
    QVector<unsigned int> ib_combined = input.index_irregular_buffer + input.index_regular_buffer;
    unsigned int offset = 0;
    for (int i = 0; i < input.vertex_buffer.length(); i++){
        //hat evtl viel redundanz

        QVector<Vertex> adj_v;//helfer
        tables.vertex_offsets.push_back(offset);
        tables.vertex_indices.push_back(i);
        offset++;

        Vertex originalVertex = input.vertex_buffer[i];

        QVector<unsigned int> d;

        for (int j = 0; j < ib_combined.length(); j++){

            //suche verweise auf vertex im indexbuffer
            if (input.vertex_buffer[ib_combined[j]].samePos(originalVertex)){
                d.push_back(j);
                unsigned int i1, i2; //indices for neighbour vertices
                switch (j % 3) {
                case 0:
                    i1 = ib_combined[j+1];
                    i2 = ib_combined[j+2];
                    break;
                case 1:
                    i1 = ib_combined[j-1];
                    i2 = ib_combined[j+1];
                    break;
                case 2:
                    i1 = ib_combined[j-2];
                    i2 = ib_combined[j-1];
                    break;
                }

                Vertex v1,v2; //neighbour vertices;
                v1 = input.vertex_buffer[i1];
                v2 = input.vertex_buffer[i2];
                //check if vertices where already used to influence this vertex
                //Note: Can't use contain. Contain uses equal (would have to implement equal, but we're only interested in the position here.
                bool found1 = false;
                bool found2 = false;
                for (int k = 0; k < adj_v.length(); k++){
                    if (adj_v[k].samePos(v1)) found1 = true;
                    if (adj_v[k].samePos(v2)) found2 = true;
                }
                if (!found1){
                    adj_v.push_back(v1);
                    tables.vertex_indices.push_back(i1);
                    offset++;
                }
                if (!found2){
                    adj_v.push_back(v2);
                    tables.vertex_indices.push_back(i2);
                    offset++;
                }
                //if one adjacent vertex is referenced by multiple different indices, only its first index will appear in the adjacent_indices list.1
            }
        }
        if (!duplicates.contains(d)) duplicates.push_back(d);
    }

    tables.vertex_offsets.push_back(offset);

    qCDebug(log_subdiv) << "Done with vertex index table. ";
    qCDebug(log_subdiv) << "Duplicates: "<<duplicates.length();
    for (int i = 0; i < duplicates.length(); i++){
        qCDebug(log_subdiv_trace) << duplicates[i];
    }
}

void Subdivision::splitRegular(Mesh *mesh) {
    QTime totalTimer;
    totalTimer.start();
    qCDebug(log_timing) << "splitRegular(mesh) started";
    Mesh::Node root = mesh->getRootNode();

    int first_mesh_index = -1;
    if (!root.getFirstMeshIndex(first_mesh_index)) {
        qCCritical(log_subdiv) << "No mesh found, aborting subdivision";
        return;
    }

    Mesh::MeshEntry *current_mesh = mesh->getMeshEntry(first_mesh_index);

    QVector<unsigned int> regular;
    QVector<unsigned int> irregular;

    findRegular(current_mesh->buffers[0]->indices_irregular, current_mesh->buffers[0]->vertices, regular, irregular);
    findRegular(current_mesh->buffers[0]->indices_regular, current_mesh->buffers[0]->vertices, regular, irregular);

    QVector<unsigned int> patches = getPatchIndexBuffer(regular, irregular,current_mesh->buffers[0]->vertices);//TODO is this the corret vertex buffer?
    current_mesh->buffers[0]->indices_regular = patches;
    current_mesh->buffers[0]->indices_irregular = irregular;
    current_mesh->buffers[0]->updateIndices();


    qCDebug(log_timing) << "splitRegular(mesh): time: " <<formatTimeMeasurement(totalTimer.elapsed());
    qCDebug(log_subdiv) << "splitRegular(mesh): regular: " << regular.length();
    qCDebug(log_subdiv) << "splitRegular(mesh): irregular: " << irregular.length();
}

/**
 * @brief Subdivision::getPatchIndexBuffer
 * Computes an index buffer for the patches. patches are in order from figure 6(a) from the paper:
 *        0     1
 *
 *     2     3     4
 *
 *  5     6     7     8
 *
 *     9     10    11
 *
 * with 367 being the original triangle.
 * Assumes that each vertex \exists only once in the vertex buffer, i.e. with regular patches, each index in the index buffer \exists 6 times.
 * (each index exists a total of 6 times in the regular and irregular index buffer combined)
 * @param ib_regular index buffer with regular triangles. Will only generate patches for these regular triangles.
 * @param ib_irregular index buffer with irregular triangles.
 * patches for regular triangles might include vertices from irregular triangles that are not part of the regular index buffer.
 * @param vb vertex buffer. We need this because the same location may appear multiple times in the vertex buffer (due to different normals / ...)
 * and simply comparing indices isn't enough.
 * @return patch index buffer with 4 times as many indices. ! if something goes wrong, the current buffer is returned and might be incomplete or empty.
 */
QVector<unsigned int> Subdivision::getPatchIndexBuffer(QVector<unsigned int> ib_regular, QVector<unsigned int> ib_irregular, QVector<Vertex> vb){
    QTime totalTimer;
    totalTimer.start();
    qCDebug(log_timing)<<"PatchIndexBuffer started";
    QVector<unsigned int> pib;
    if (ib_regular.size() == 0) return pib;
    QTime tempTimer;
    tempTimer.start();
    QVector<unsigned int> ib_combined = ib_regular + ib_irregular; //DO NOT swap regular and irregular! iteration over ib_combined starting from index in ib_regular
    qCDebug(log_timing) << "ib_combined:" << formatTimeMeasurement(tempTimer.elapsed());
    QMap<Triangle,QVector<Triangle>> adj;
   // qDebug()<<"combined lenght"<<ib_combined.length();
    for (int i = 0; i < ib_combined.length() - 2; i+=3){
     //   qDebug()<<"Index"<<i;
        unsigned int i3,i6,i7;

        //alles gegen den uzs
        i3 = ib_combined[i];
        i6 = ib_combined[i+1];
        i7 = ib_combined[i+2];


        if(log_subdiv_trace().isDebugEnabled()){
            if (i < ib_regular.length()){
                Vertex v3 = vb[i3];
                Vertex v6 = vb[i6];
                Vertex v7 = vb[i7];

                int count3 = 0; //counts other triangles with this vertex
                int count6 = 0;
                int count7 = 0;
                for (int j = 0; j < ib_combined.length(); j++){
                    if (j != i && j != i+1 && j != i+2){
                        //for debugging // checking if patch is regular.
                        if (vb[ib_combined[j]].samePos(v3)) count3++;
                        if (vb[ib_combined[j]].samePos(v6)) count6++;
                        if (vb[ib_combined[j]].samePos(v7)) count7++;
                    }
                }
                if (count3 != 5 || count6 != 5 || count7 != 5){
                    qCWarning(log_subdiv) << "Counts wrong! 3: "<< count3 <<", 6: " << count6 << ", 7: " << count7 << ".\nEither this patch is not regular, or you didn't pass all neighboring triangles.";
                }

            }
        }

        Triangle t367(&vb,i3,i6,i7);

        for (int j = i+3; j < ib_combined.length()-2; j += 3){
            unsigned int j0 = ib_combined[j];
            unsigned int j1 = ib_combined[j+1];
            unsigned int j2 = ib_combined[j+2];

            Triangle tj(&vb,j0,j1,j2);

            if (t367.hasSharedEdge(tj)){
                QVector<Triangle> neighbors = adj.value(tj,QVector<Triangle>());
                neighbors.push_back(t367);
                adj.insert(tj,neighbors);
                neighbors = adj.value(t367,QVector<Triangle>());
                if (neighbors.length() == 3) break;
                neighbors.push_back(tj);
                adj.insert(t367,neighbors);
            }
        }

        if (adj.value(t367,QVector<Triangle>()).size() != 3){
            qCDebug(log_subdiv_trace)<<"no 3 adjacent triangles found";
        }
    }

    for(int i = 0; i < ib_regular.length()-2; i+=3){
        unsigned int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, i10, i11;
        i3 = ib_regular[i];
        i6 = ib_regular[i+1];
        i7 = ib_regular[i+2];
        Triangle ti(&vb,i3,i6,i7);
        bool found0, found1, found5, found8, found9, found11;
        QVector<Triangle> neighbors = adj.value(ti,QVector<Triangle>());
        for (int j = 0; j < neighbors.length(); j++){
            Triangle::Edge a,b;
            Triangle tj = neighbors[j];
            ti.get_shared_edge(tj,a,b);

            Triangle::Edge c,d;
            QVector<Triangle> an = adj.value(tj,QVector<Triangle>());
            if (a.name == Triangle::Edge::Name::uv){
                i2 = b.c;

                for(int k = 0; k < an.length(); k++){
                    tj.get_shared_edge(an[k],c,d);
                    if (tj.next_ctrclockwise(b,c)){
                        i0 = d.c;
                        found0 = true;
                    }
                    if (tj.next_clockwise(b,c)){
                        i5 = d.c;
                        found5 = true;
                    }
                }
            } else if (a.name == Triangle::Edge::Name::vw){
                i10 = b.c;
                for (int k = 0; k < an.length(); k++){
                    tj.get_shared_edge(an[k],c,d);
                    if (tj.next_ctrclockwise(b,c)){
                        i9 = d.c;
                        found9 = true;
                    }
                    if (tj.next_clockwise(b,c)){
                        i11 = d.c;
                        found11 = true;
                    }
                }
            } else if (a.name == Triangle::Edge::Name::wu){
                i4 = b.c;
                for (int k = 0; k < an.length(); k++){
                    tj.get_shared_edge(an[k],c,d);
                    if (tj.next_ctrclockwise(b,c)){
                        i8 = d.c;
                        found8 = true;
                    }
                    if (tj.next_clockwise(b,c)){
                        i1 = d.c;
                        found1 = true;
                    }
                }
            }
        }
        if (!(found0 && found1 && found5 && found9 && found8 && found11)){

                 qCWarning(log_subdiv) << "Couldnt find some neighbour at i= " << i;
                 qCWarning(log_subdiv) << "i3: " << i3 << ", i6: " << i6 << " i7: " << i7 << " i2 " << i2;


                 for (int k = 0; k < ib_combined.length() - 2; k+= 3){
                     if (ib_combined[k] == i3 || ib_combined[k+1] == i3 || ib_combined[k+2] == i3){
                         qCWarning(log_subdiv) << "i3 adjacent: " << ib_combined[k] << ", " << ib_combined[k+1] << ", " << ib_combined[k+2];
                     }
                 }

                 qCWarning(log_subdiv) << "found: 0  1  5  9  8  11  \n" << found0 << found1 << found5 << found9 << found8 << found11;
                 qCWarning(log_subdiv) << "Abort computing patch index buffer.";
                 return pib;
         }
             //else qCWarning(log_subdiv) << "found all neighbours for patch.";
         pib.push_back(i0);
         pib.push_back(i1);
         pib.push_back(i2);
         pib.push_back(i3);
         pib.push_back(i4);
         pib.push_back(i5);
         pib.push_back(i6);
         pib.push_back(i7);
         pib.push_back(i8);
         pib.push_back(i9);
         pib.push_back(i10);
         pib.push_back(i11);
    }

    qCDebug(log_timing)<<"PatchIndexBuffer done:"<<formatTimeMeasurement(totalTimer.elapsed());
    return pib;
}

/**
 * Turns a patch index buffer into an index buffer of triangles.
 */
QVector<unsigned int> Subdivision::patchIBToTriangleIB(QVector<unsigned int> ib) {
    QVector<unsigned int> new_ib;
    for (int i = 0; i < ib.length(); i += 12) {
        new_ib.push_back(ib[i+3]);
        new_ib.push_back(ib[i+6]);
        new_ib.push_back(ib[i+7]);
    }
    return new_ib;
}

/**
 * Generates index buffers containing all regular and all irregular triangles in the input index buffer.
 *
 * @param index_buffer Index buffer describing triangles.
 * @param vertex_buffer Vertices used by index_buffer.
 * @param regular All regular triangles are appended to this index buffer.
 * @param irregular All irregular triangles are appended to this index buffer.
 */
void Subdivision::findRegular(QVector<unsigned int> index_buffer, QVector<Vertex> vertex_buffer, QVector<unsigned int> &regular, QVector<unsigned int> &irregular) {
    QTime totalTimer;
    totalTimer.start();
    qCDebug(log_timing)<<"Find Regular started";
    for (int i = 0; i < index_buffer.length()-2; i += 3){
        Vertex vx = vertex_buffer[index_buffer[i]];
        Vertex vy = vertex_buffer[index_buffer[i+1]];
        Vertex vz = vertex_buffer[index_buffer[i+2]];

        int countx = 0;
        int county = 0;
        int countz = 0;
        for (int j = 0; j < index_buffer.length(); j++){
            if (j != i && j != i+1 && j != i+2){
                if (vx.samePos(vertex_buffer[index_buffer[j]])){
                    countx++;
                }
                if (vy.samePos(vertex_buffer[index_buffer[j]])){
                    county++;
                }
                if (vz.samePos(vertex_buffer[index_buffer[j]])){
                    countz++;
                }
            }
        }
        if (countx == 5 && county == 5 && countz == 5){
            regular.push_back(index_buffer[i]);
            regular.push_back(index_buffer[i+1]);
            regular.push_back(index_buffer[i+2]);
        } else {
            irregular.push_back(index_buffer[i]);
            irregular.push_back(index_buffer[i+1]);
            irregular.push_back(index_buffer[i+2]);
        }
    }
    qCDebug(log_timing)<<"Find Regular done:"<<formatTimeMeasurement(totalTimer.elapsed());
}

bool Subdivision::matchAndCompleteTriangleIndices(unsigned int sx, unsigned int sy, unsigned int sz, unsigned int tx, unsigned int ty, unsigned int &tz){
    if (sx == tx && sy == ty){
        tz = sz;
        return true;
    }
    if (sy == tx && sz == ty){
        tz = sx;
        return true;
    }
    if (sz == tx && sx == ty){
        tz = sy;
        return true;
    }
    return false;
}


bool Subdivision::matchAndCompleteTriangle(unsigned int sx, unsigned int sy, unsigned int sz,unsigned int tx, unsigned int ty, unsigned int &tz, QVector<Vertex> vb){
    if (vb[sx].samePos(vb[tx]) && vb[sy].samePos(vb[ty])){
        tz = sz;
        return true;
    }
    if (vb[sy].samePos(vb[tx]) && vb[sz].samePos(vb[ty])){
        tz = sx;
        return true;
    }
    if (vb[sz].samePos(vb[tx]) && vb[sx].samePos(vb[ty])){
        tz = sy;
        return true;
    }
    return false;

}


Subdivision::Result Subdivision::runShader(Input input, Tables &tables) {
    qCDebug(log_subdiv) << "Running compute shader";

    Result result;

    // Create an input buffer for the vertex indices
    GLuint vertex_indices_handle;
    f->glGenBuffers(1, &vertex_indices_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, vertex_indices_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, tables.vertex_indices.size() * sizeof(GLuint), tables.vertex_indices.data(), GL_DYNAMIC_DRAW);

    // Create an input buffer for the vertex offsets
    GLuint vertex_offsets_handle;
    f->glGenBuffers(1, &vertex_offsets_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, vertex_offsets_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, tables.vertex_offsets.size() * sizeof(GLuint), tables.vertex_offsets.data(), GL_DYNAMIC_DRAW);

    // Create an input buffer for the edge indices
    GLuint edge_indices_handle;
    f->glGenBuffers(1, &edge_indices_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, edge_indices_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, tables.edge_indices.size() * sizeof(GLuint), tables.edge_indices.data(), GL_DYNAMIC_DRAW);

    // Create an output buffer
    GLuint output_handle;
    f->glGenBuffers(1, &output_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, output_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, (input.vertex_buffer.size() + tables.edge_indices.size() / 4) * sizeof(Vertex), NULL, GL_DYNAMIC_DRAW);
    result.vb_handle = output_handle;

    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0);
    QTime timer;
    timer.start();
    runVertexShader(input.vertex_buffer.size(), input.vb_handle, vertex_indices_handle, vertex_offsets_handle, output_handle);
    int vertexTime = timer.elapsed();
    timer.restart();
    int edgeOffset = input.vertex_buffer.size();
    runEdgeShader(tables.edge_indices.size() / 4, input.vb_handle, edge_indices_handle, output_handle, edgeOffset);
    int edgeTime = timer.elapsed();
    timer.restart();

    // Map the output buffer so we can read the results
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, output_handle);
    Vertex *ptr;
    ptr = (Vertex *) f->glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_WRITE_ONLY);

    qCDebug(log_subdiv_trace) << "New vertices:";
    for (int i = 0; i < input.vertex_buffer.size(); i++) {
        qCDebug(log_subdiv_trace) << ptr[i].pos;
        result.vertex_buffer.push_back(ptr[i]);
    }
    qCDebug(log_subdiv_trace) << "New edge points:";
    for (int i = 0; i < tables.edge_indices.size() / 4; i++) {
        qCDebug(log_subdiv_trace) << ptr[edgeOffset + i].pos;
        result.vertex_buffer.push_back(ptr[edgeOffset + i]);
    }
    int bufferbildTime = timer.elapsed();

    f->glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);

    // Delete the buffers the free the resources
    f->glDeleteBuffers(1, &vertex_indices_handle);
    f->glDeleteBuffers(1, &vertex_offsets_handle);
    f->glDeleteBuffers(1, &edge_indices_handle);

    qCDebug(log_timing)<<"Compute Shader done. Vertexshader"<<vertexTime<<"ms; EdgeShader"<<edgeTime<<" BufferToCPU"<<bufferbildTime<<"ms";
    return result;
}

void Subdivision::runVertexShader(GLuint size, GLuint vb_handle, GLuint vertex_indices_handle, GLuint vertex_offsets_handle, GLuint output_handle) {
    vertexShader->bind();

    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, vb_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, vertex_indices_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, vertex_offsets_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, output_handle);

    // Run the shader
    f->glDispatchCompute(size, 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, 0);

    vertexShader->release();
}

void Subdivision::runEdgeShader(GLuint size, GLuint vb_handle, GLuint edge_indices_handle, GLuint output_handle, GLuint offset) {
    edgeShader->bind();

    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, vb_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, edge_indices_handle);
    f->glBindBufferRange(GL_SHADER_STORAGE_BUFFER, 2, output_handle, offset * sizeof(Vertex), size * sizeof(Vertex));

    // Run the shader
    f->glDispatchCompute(size, 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, 0);

    edgeShader->release();
}
